/****************************************************************************
** $Id: qt/src/kernel/qwsqpedecoration_qws.cpp   2.2.3   edited 2000-12-12 $
**
** Implementation of Qt/Embedded window manager decorations
**
** Created : 000101
**
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of the kernel module of the Qt GUI Toolkit.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding valid Qt Enterprise Edition or Qt Professional Edition
** licenses for Qt/Embedded may use this file in accordance with the
** Qt Embedded Commercial License Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/pricing.html or email sales@trolltech.com for
**   information about Qt Commercial License Agreements.
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include <qapplication.h>
#include <qwidget.h>
#include <qpainter.h>
#include "qwsqpedecoration_qws.h"

#ifndef QT_NO_QWS_QPE_WM_STYLE

#ifndef QT_NO_IMAGEIO_XPM

/* XPM */
static const char * const qpe_close_xpm[] = {
"16 16 3 1",
"       c None",
".      c #FFFFFF",
"+      c #000000",
"                ",
"                ",
"      .....     ",
"    ..+++++..   ",
"   .+++++++++.  ",
"   .+..+++..+.  ",
"  .++...+...++. ",
"  .+++.....+++. ",
"  .++++...++++. ",
"  .+++.....+++. ",
"  .++...+...++. ",
"   .+..+++..+.  ",
"   .+++++++++.  ",
"    ..+++++..   ",
"      .....     ",
"                "};

/* XPM */
static const char * const qpe_accept_xpm[] = {
"16 16 3 1",
"       c None",
".      c #FFFFFF",
"+      c #000000",
"                ",
"                ",
"      .....     ",
"    ..+++++..   ",
"   .+++++++++.  ",
"   .+++++++++.  ",
"  .+++++++..++. ",
"  .++.+++...++. ",
"  .+...+...+++. ",
"  .+......++++. ",
"  .++....+++++. ",
"   .++..+++++.  ",
"   .+++++++++.  ",
"    ..+++++..   ",
"      .....     ",
"                "};

#endif // QT_NO_IMAGEIO_XPM


QWSQPEDecoration::QWSQPEDecoration()
    : QWSDefaultDecoration()
{
}

QWSQPEDecoration::~QWSQPEDecoration()
{
}

const char **QWSQPEDecoration::menuPixmap()
{
    return (const char **)0;
}

const char **QWSQPEDecoration::closePixmap()
{
    return (const char **)qpe_close_xpm;
}

const char **QWSQPEDecoration::minimizePixmap()
{
    return (const char **)qpe_accept_xpm;
}

const char **QWSQPEDecoration::maximizePixmap()
{
    return (const char **)0;
}

const char **QWSQPEDecoration::normalizePixmap()
{
    return (const char **)0;
}

int QWSQPEDecoration::getTitleHeight(const QWidget *)
{
    return 15;
}

/*
    If rect is empty, no frame is added. (a hack, really)
*/
QRegion QWSQPEDecoration::region(const QWidget *widget, const QRect &rect, QWSDecoration::Region type)
{
    int titleHeight = getTitleHeight(widget);
//    int titleWidth = getTitleWidth(widget);
//    int bw = rect.isEmpty() ? 0 : BORDER_WIDTH;
    QRegion region;

/*
#ifdef QPE_WM_LOOK_AND_FEEL
    int bw = rect.isEmpty() ? 0 : BORDER_WIDTH;
    int bbw = rect.isEmpty() ? 0 : BOTTOM_BORDER_WIDTH;

    switch (type) {
        case All: {
                region = QRect( rect.left() - bw,
                            rect.top() - bw - titleHeight,
                            rect.width() + 2 * bw,
                            rect.height() + bw + bbw + titleHeight);
                region += region(w, rect, Title);
                region += region(w, rect, Close);
                region += region(w, rect, Minimize);
                region += region(w, rect, BottomRight);
                region -= rect;
            }
            break;

        case Title: {
                QRect r(rect.left() - bw,
                        rect.top() - titleHeight,
                        rect.width() - titleHeight + bw
                         - (w->inherits( "QDialog" ) ? titleHeight : 0),
                        titleHeight);
                if ( r.width() > 0 )
                    region = r;
            }
            break;

        case Bottom: {
                QRect r1(rect.left() - bw,
                        rect.bottom() + 1,
                        rect.width() - CORNER_GRAB + bw,
                        bbw);

                if ( r1.left() < rect.left() )
                    r1.setLeft( rect.left() );
                region = r1;
            }
            break;

        case BottomRight: {
                QRect r1(rect.right() - CORNER_GRAB,
                        rect.bottom() + 1,
                        CORNER_GRAB + bw,
                        bbw);

                if ( r1.left() < rect.left() )
                    r1.setLeft( rect.left() );
                region = r1;
            }
            break;

        case Minimize:
            if ( w->inherits( "QDialog" ) ) {
                QRect r(rect.right() - 2*titleHeight,
                        rect.top() - titleHeight,
                        titleHeight, titleHeight );
                if (r.left() > rect.left() + titleHeight)
                    region = r;
            }
            break;
    }
#endif
*/

    switch (type) {
	case Menu:
	case Maximize:
	    break;
	case Minimize: {
            if ( widget->inherits( "QDialog" ) ) {
		QRect r(rect.right() - 2*titleHeight, rect.top() - titleHeight - 2,
		    titleHeight, titleHeight);
		if (r.left() > rect.left() + titleHeight)
		    region = r;
	    }
	    break;
	}
	case Close: {
		QRect r;
		r = QRect(rect.right() - titleHeight, rect.top() - titleHeight - 2,
		    titleHeight, titleHeight);
		if (r.left() > rect.left() + titleHeight)
		    region = r;
	    break;
	}
	case All:
	case Title:
	case Top:
	case Left:
	case Right:
	case Bottom:
	case TopLeft:
	case TopRight:
	case BottomLeft:
	case BottomRight:
	default:
	    region = QWSDefaultDecoration::region(widget, rect, type);
	    break;
    }

    return region;
}

void QWSQPEDecoration::paint(QPainter *painter, const QWidget *widget)
{
#ifndef QT_NO_STYLE
    QStyle &style = QApplication::style();
#endif

    int titleWidth = getTitleWidth(widget);
    int titleHeight = getTitleHeight(widget);

    QRect rect(widget->rect());

    // Border rect
    QRect br( rect.left() - BORDER_WIDTH,
                rect.top() - BORDER_WIDTH - titleHeight,
                rect.width() + 2 * BORDER_WIDTH,
                rect.height() + BORDER_WIDTH + BOTTOM_BORDER_WIDTH + titleHeight );

    // title bar rect
    QRect tr;

    tr = QRect( rect.left(), rect.top() - titleHeight, rect.width(), titleHeight );

    QRegion oldClip = painter->clipRegion();
    painter->setClipRegion( oldClip - QRegion( tr ) );	// reduce flicker

#ifndef QT_NO_PALETTE
    // const QColorGroup &cg = QApplication::palette().active();
    const QColorGroup &cg = widget->palette().active();

#if !defined(QT_NO_STYLE)
    style.drawPanel(painter, br.x(), br.y(), br.width(),
		    br.height() - 4, cg, FALSE, 2,
		    &cg.brush(QColorGroup::Background));
#elif !defined(QT_NO_DRAWUTIL)
    qDrawWinPanel(painter, br.x(), br.y(), br.width(),
		  br.height() - 4, cg, FALSE,
		  &cg.brush(QColorGroup::Background));
#endif

    painter->setClipRegion( oldClip );

    if (titleWidth > 0) {
	QBrush titleBrush;
	QPen   titlePen;
	int    titleLeft = titleHeight + 4;

	if (widget == qApp->activeWindow()) {
	    titleBrush = cg.brush(QColorGroup::Highlight);
	    titlePen   = cg.color(QColorGroup::HighlightedText);
	} else {
	    titleBrush = cg.brush(QColorGroup::Background);
	    titlePen   = cg.color(QColorGroup::Text);
	}

#define CLAMP(x, y)	    ( ((x) > (y)) ? (y) : (x) )

	    titleLeft = rect.left() + 5;
	    painter->setPen( cg.midlight() );
	    painter->drawLine( rect.left() - BORDER_WIDTH + 2,
				rect.bottom() + 1, rect.right() + BORDER_WIDTH - 2,
				rect.bottom() + 1 );
	    style.drawSeparator( painter, rect.right() - CORNER_GRAB,
			    rect.bottom() + 1, rect.right() - CORNER_GRAB,
			    rect.bottom() + BOTTOM_BORDER_WIDTH,
			    cg, TRUE, 1, 0 );
	    painter->fillRect(rect.left() - 2,
			    rect.top() - titleHeight - 2,
			    rect.width() + 3, titleHeight + 2,
			    titleBrush);

	painter->setPen(titlePen);
	painter->setFont(widget->font());
	painter->drawText( titleLeft, -titleHeight,
	 		rect.width() - titleHeight - 10, titleHeight-1,
			QPainter::AlignVCenter, widget->caption());
    }

#endif //QT_NO_PALETTE

}

void QWSQPEDecoration::paintButton(QPainter *painter, const QWidget *w,
			QWSDecoration::Region type, int state)
{
#ifndef QT_NO_PALETTE
#ifndef QT_NO_STYLE
//    QStyle &style = QApplication::style();
#endif
    const QColorGroup &cg = w->palette().active();

    QRect brect(region(w, w->rect(), type).boundingRect());

    int xoff=2;
    int yoff=2;

    const QPixmap *pm=pixmapFor(w,type,state & QWSButton::On, xoff, yoff);

	switch (type) {
	    case Close:
	    case Menu:
	    case Minimize:
		break;
	    default:
		return;
	}

	QBrush titleBrush;

	if (w == qApp->activeWindow()) {
	    titleBrush = cg.brush(QColorGroup::Highlight);
	} else {
	    titleBrush = cg.brush(QColorGroup::Background);
	}

	painter->fillRect( brect.x(), brect.y(), brect.width()+1,
		    brect.height()+1, titleBrush );
	if ((state & QWSButton::MouseOver) && (state & QWSButton::Clicked)) {
	    if (pm) painter->drawPixmap(brect.x()+1, brect.y()+1, *pm);
	} else {
	    if (pm) painter->drawPixmap(brect.x(), brect.y(), *pm);
	}

#endif

}

#endif // QT_NO_QWS_QPE_WM_STYLE
